//*****************************************************************************
// file:        btfbase.frag
// version:     1.0
// description: BTFBASE GPU fragment shader, LDR and HDR version
// author:      Jiri Filip, UTIA AS CR
// date:        28/01/2010
//*****************************************************************************
/*
COPYRIGHT ======================================================================

This source code is release under Creative Commons licence
(Attribution-NonCommercial 3.0 Unported) specified here:

http://creativecommons.org/licenses/by-nc/3.0/

You are free:
    * to Share  to copy, distribute and transmit the work
    * to Remix  to adapt the work
Under the following conditions:
    * Attribution  You must attribute the work in the manner specified by the
      author or licensor (but not in any way that suggests that they endorse you
      or your use of the work).

      Attribute this work:
      ------------------------------------------------
      Bidirectional Texture Function Compression based on Multi-Level Vector 
      Quantization
      as the paper written by
          Vlastimil Havran 	Jiri Filip 	Karol Myszkowski

      and published in Computer Graphics Forum, Volume 29, Issue 1, 
      pages 175-190, March 2010, DOI: 10.1111/j.1467-8659.2009.01585.x

      Project page:
        http://www.cgg.cvut.cz/members/havran/btfbase/
      ------------------------------------------------
     
    * Noncommercial  You may not use this work for commercial purposes.

With the understanding that:
    * Waiver  Any of the above conditions can be waived if you get permission 
      from the copyright holder.
    * Public Domain  Where the work or any of its elements is in the public 
      domain under applicable law, that status is in no way affected by 
      the license.
    * Other Rights  In no way are any of the following rights affected by the 
      license:
          o Your fair dealing or fair use rights, or other applicable copyright 
            exceptions and limitations;
          o The author's moral rights;
          o Rights other persons may have either in the work itself or in how 
            the work is used, such as publicity or privacy rights.
    * Notice  For any reuse or distribution, you must make clear to others 
         the license terms of this work. The best way to do this is with a link 
         to this web page.
=============================================================================
PIXEL EVALUATION PIPELINE:
        phi        beta
xy  theta     alpha    ab
                 ->P1        \
             ->P2            |
            |    ->P1        |
            |    ->P1        |
            |->P2            |
            |    ->P1        | 
          ->M                |
         |  |    ->I1->C(a)  |
         |  |   |   `->C(b)  > TREE
         |  |->I2            |
         |  |    ->I1->C(a)  |
         |  |       `->C(b)  |
         |  |    ->I1->C(a)  |
         |  |   |   `->C(b)  | 
         |   ->I2            |
         |       ->I1->C(a)  |
         |          `->C(b)  /
      ->P3
     |    ->TREE
P6->P4
     |    ->TREE
      ->P3
          ->TREE
No. of evaluations (texture lookups) of individual codebooks (Fig.8 in paper).
P6  P4  P3  M P2/I2 P1/I1 C    ie. total: 40 integer (16bit) texture reads     
1   1   2   4  8/8  16/16 32              60 float   (32bit) texture reads

BTFBASE data ordering in textures

   |        indices - GL_RGBA16        |   GL_RGBA16(32)F_ARB   |
   | TI16_1 | TI16_2 | TI16_3 | TI16_4 | TF32_scales| TF32_data |
   |--------|--------|--------|--------|------------|-----------| 
  R| P6.x   | P3.x   | M.x    | I2.x   |   P6       |   P1      |
  G| P6.y   | P3.y   | M.y    | I2.y   |   P4       |   C       |
  B| P4.x   | P2.x   | empty  | I1.x   |   P3       |   empty   |
  A| P4.y   | P2.y   | empty  | I1.y   |   P2       |   empty   | 
=============================================================================*/

uniform samplerCube cubemap_ba;   // cubemaps with indices and weights corresponding 
uniform samplerCube cubemap_tv;   // to alpha,beta, theta_v,phi_v angles

// INT16bit textures containing codebooks indices (see the table above)
uniform sampler2D TI16_1;  
uniform sampler2D TI16_2;  
uniform sampler2D TI16_3;  
uniform sampler2D TI16_4;  

// FLOAT32bit textures containing codebooks scales/data (see the table above)
uniform sampler2D TF32_scales;  
uniform sampler2D TF32_data;  

varying vec3 T; // tangent
varying vec3 B; // binormal
varying vec3 N; // normal

varying vec3 L; // normalized light vector
varying vec3 C; // normalized light vector

uniform float texSize;        // parametric textures size - physical (typically 1024,2048)
uniform float texMapSizeX;    // mapped texture X resolution (arbitrary)
uniform float texMapSizeY;    // mapped texture Y resolution (arbitrary)
uniform float pixStep;        // 1/texSize = 1/2048 = 0.00048828125;
uniform float BTFheight;      // Size of texture
uniform float BTFwidth;
uniform float texCutX;        // BTFwidth/texSize 
uniform float texCutY;        // BTFheight/texSize

uniform float SlicesPerPhi;   // 16 ??? why not general!!!!

uniform float hdrMultiplier;  // Multiplier for HDR materials     

//########################################
// YCbCr -> RGB conversion in pixel shader
//########################################

vec3 YCbCrtoRGB(in float Y, in float Cb, in float Cr)
{
  vec3 resRGB;
  float C;
  C = Y * 1.1643828;
  resRGB[0] = C +                   Cr * 1.5960273  - 222.921;
  resRGB[1] = C - Cb * 0.39176172 - Cr * 0.81296875 + 135.576;
  resRGB[2] = C + Cb * 2.0172344                    - 276.836;

  return resRGB/255.0;

}//--- YCbCrtoRGB ----------------------------------------- 


vec3 LogLuv2RGB(in float L, in float U, in float V)
{
  float X,Y,Z;
  if (L < 98.381)
     Y = 0.056968 * L;
  else
    if (L < 1204.7)
  	  Y = 7.3014e-30 * pow(L + 884.17, 9.9872);
    else
	  Y = 32.994 * exp(0.0047811 * L);
  X = 0.0;
  Z = 0.0;
  if (V > 0.0) {
    X = 2.25 * U/V * Y;
    Z = Y * (1230.0/V - 5.0) - X/3.0; 
  }
  
  vec3 resRGB; 
  resRGB.x = X *  3.2406 + Y * -1.5372 + Z * -0.4986;
  resRGB.y = X * -0.9689 + Y *  1.8758 + Z *  0.0415;
  resRGB.z = X *  0.0557 + Y * -0.2040 + Z *  1.0570;

  if (resRGB.x < 0.0) resRGB.x = 0.0;
  if (resRGB.y < 0.0) resRGB.y = 0.0;
  if (resRGB.z < 0.0) resRGB.z = 0.0;

  // 11/01/2009 - in fact, this should not be here - but it seems
  // that all UBO HDR images are gamma corrected  - Vlastimil Havran
  if (resRGB.x > 0.0031308) resRGB.x = 1.055 * pow(resRGB.x, 1.0/2.4) - 0.055;
  else               resRGB.x = 12.92 * resRGB.x;
  if (resRGB.y > 0.0031308) resRGB.y = 1.055 * pow(resRGB.y, 1.0/2.4) - 0.055;
  else               resRGB.y = 12.92 * resRGB.y;
  if (resRGB.z > 0.0031308) resRGB.z = 1.055 * pow(resRGB.z, 1.0/2.4) - 0.055;
  else               resRGB.z = 12.92 * resRGB.z;

  resRGB *= hdrMultiplier;

  return resRGB;
}//--- LogLuv2RGB -------------------------------------------

void main()
{
  // cubemap index vectors (pointing to light and camera) computation =================
  // light vector computation and transformation into tangent space
  vec3 lightDir;
  lightDir.x = dot(L, T);
  lightDir.y = dot(L, B);
  lightDir.z = dot(L, N);
  //lightDir = normalize(lightDir);
      
  // camera vector computation and transformation into tangent space
  vec3 cameraDir;
  cameraDir.x = dot(C, T);
  cameraDir.y = dot(C, B);
  cameraDir.z = dot(C, N);
  //cameraDir = normalize(cameraDir);

  const float PI = 3.14159265359;
  // pick up view indices and weigths from cubemaps ====================================
  float it = 255.0 * textureCube(cubemap_tv, cameraDir).r;
  float ip = 255.0 * (textureCube(cubemap_tv, cameraDir).g + 3e-7);
  float wt = textureCube(cubemap_tv,cameraDir).b;
  float wp = textureCube(cubemap_tv,cameraDir).a;

  float theta_i = acos(lightDir.z);
  float theta_v = acos(cameraDir.z);

  // removing occluded and/or non-lighted pixels 
  float halfPI = PI/2.0;
  if(theta_i>halfPI || theta_v>halfPI)
    gl_FragColor = vec4(0.0,0.0,0.0,1.0);
  else
  {// if pixel visible and illuminated then continue with its processing...

  const float twoPI = 2.0 * PI;
  float phi_i = atan(lightDir.y, lightDir.x);
  phi_i += PI; // convert from (-PI,PI) to (0,2PI)
  phi_i = mod(phi_i, twoPI);

  float phi_v = atan(cameraDir.y, cameraDir.x);
  phi_v += PI; // convert from (-PI,PI) to (0,2PI)
  phi_v = mod(phi_v, twoPI);

  phi_v = twoPI - phi_v;
  phi_v = mod(phi_v, twoPI);

  // recompute from clockwise to anti-clockwise phi_v notation: (360.f - phi_i) 
  // rotation of onion-sliced parametrization to be perpendicular to phi_v of viewing angle: - (90.f + phi_v)
  phi_i = (twoPI - phi_i) - (0.5 * PI + phi_v);
  phi_i = mod(phi_i, twoPI);
  if(phi_i < 0.0)
    phi_i += twoPI;

  // back to vector
  vec3 newLightDir;
  float xy = sin(theta_i);
  newLightDir.x = xy * cos(phi_i);
  newLightDir.y = xy * sin(phi_i);
  newLightDir.z = cos(theta_i);
  newLightDir = normalize(newLightDir);
  // pickup illum. indices and weigths from cubemaps for new vector ==================
  float ib = 255.0 * textureCube(cubemap_ba, newLightDir ).r;
  float ia = 255.0 * textureCube(cubemap_ba, newLightDir ).g;
  float wb = textureCube(cubemap_ba,newLightDir).b;           
  float wa = textureCube(cubemap_ba,newLightDir).a;

  // compute inverse weights
  float i_wa = 1.0 - wa;
  float i_wt = 1.0 - wt;
  float i_wp = 1.0 - wp;

  // repetitive texture mapping =======================================================
  vec2 UV;
  UV.x = texMapSizeX*gl_TexCoord[0].s;
  UV.y = texMapSizeY*gl_TexCoord[0].t;
  float aux;
  aux = UV.x/BTFwidth;
  UV.x = texCutX*(aux-floor(aux));
  aux = UV.y/BTFheight;
  UV.y = texCutY*(aux-floor(aux));

  //##############################
  // INDEX AND SCALES LOOKUPS
  //##############################

  float shiftPix = 1.0/(2.0*texSize); // half of pix. width
  float normUV   = 65534.999999/texSize;     // 65535 * pixSize = 65535 / texsize (1024,2048) 

  //PDF6D lookup ==========================================
  float shiftP4 = -0.00013/(2.0*texSize);
  vec2  iP4 = shiftP4 + normUV* texture2D(TI16_1, UV).rg;  // P6index[y][x]-1;

  float cP4 = texture2D(TF32_scales, UV).r;      // P6scale[y][x];
  
  //PDF4D lookups ==========================================
  float ip2 = mod(ip + 1.0, float(SlicesPerPhi));
  
  vec2 iP3_1, iP3_2;
  UV.y = iP4.y;
  UV.x = iP4.x + ip*pixStep;
  float shiftP3 = 1.0/(2.0*texSize);
  float normUV3 = 65534.99999/texSize;
  iP3_1 = shiftP3+normUV3* texture2D(TI16_1, UV).ba;     // P4index[iP4][ip];
  float cP3_1 = i_wp * texture2D(TF32_scales, UV).g;     // P4scale[iP4][ip]*(1.0-wp);
  UV.x = iP4.x + ip2*pixStep;
  iP3_2 = shiftP3+normUV3* texture2D(TI16_1, UV).ba;     // P4index[iP4][ip2];
  float cP3_2 =   wp * texture2D(TF32_scales, UV).g;     // P4scale[iP4][ip2]*wp;

  //PDF3D lookups ==========================================
  vec2  iM_11,iM_21, iM_12,iM_22; 
  float cM_11,cM_21, cM_12,cM_22;
  aux = it*pixStep;

  iP3_1.x += aux;
  iM_11 = shiftPix+normUV* texture2D(TI16_2, iP3_1).rg;  // P3index[iP3_1][it];
  cM_11 = texture2D(TF32_scales, iP3_1).b * i_wt;        // P3scale[iP3_1][it]*1_wt;
  iP3_1.x += pixStep;
  iM_21 = shiftPix+normUV* texture2D(TI16_2, iP3_1).rg;  // P3index[iP3_1][it+1];
  cM_21 = texture2D(TF32_scales, iP3_1).b * wt;          // P3scale[iP3_1][it+1]*wt;

  iP3_2.x += aux;
  iM_12 = shiftPix+normUV* texture2D(TI16_2, iP3_2).rg;  // P3index[iP3_2][it];
  cM_12 = texture2D(TF32_scales, iP3_2).b * i_wt;        // P3scale[iP3_2][it]*1_wt;
  iP3_2.x += pixStep;
  iM_22 = shiftPix+normUV* texture2D(TI16_2, iP3_2).rg;  // P3index[iP3_2][it+1];
  cM_22 = texture2D(TF32_scales, iP3_2).b * wt;          // P3scale[iP3_2][it+1]*wt;

  // M lookups =============================================
  vec2 iP2_0, iI2_0, iP2_1,iI2_1, iP2_2,iI2_2, iP2_3,iI2_3;

  iP2_0 = shiftPix+normUV* texture2D(TI16_3, iM_11).rg;  // Mindex[iM_1[0]][0];
  iM_11.x += pixStep;              
  iI2_0 = shiftPix+normUV* texture2D(TI16_3, iM_11).rg;  // Mindex[iM_1[0]][1];

  iP2_1 = shiftPix+normUV* texture2D(TI16_3, iM_21).rg;  // Mindex[iM_2[0]][0];
  iM_21.x += pixStep;              
  iI2_1 = shiftPix+normUV* texture2D(TI16_3, iM_21).rg;  // Mindex[iM_2[0]][1];
  
  iP2_2 = shiftPix+normUV* texture2D(TI16_3, iM_12).rg;  // Mindex[iM_1[1]][0];
  iM_12.x += pixStep;              
  iI2_2 = shiftPix+normUV* texture2D(TI16_3, iM_12).rg;  // Mindex[iM_1[1]][1];

  iP2_3 = shiftPix+normUV* texture2D(TI16_3, iM_22).rg;  // Mindex[iM_2[1]][0];
  iM_22.x += pixStep;              
  iI2_3 = shiftPix+normUV* texture2D(TI16_3, iM_22).rg;  // Mindex[iM_2[1]][1];

  //PDF2D lookups ==========================================
  vec2  iP1_10,iP1_20, iP1_11,iP1_21, iP1_12,iP1_22, iP1_13,iP1_23;   
  float cP1_10,cP1_20, cP1_11,cP1_21, cP1_12,cP1_22, cP1_13,cP1_23;
  aux = ia*pixStep;
  
  iP2_0.x += aux;
  iP1_10 = shiftPix+normUV* texture2D(TI16_2, iP2_0).ba; // P2index[iP2[0]][ia];
  cP1_10 = texture2D(TF32_scales, iP2_0).a * i_wa;       // P2scale[iP2[0]][ia]*1_wa;
  iP2_0.x += pixStep;
  iP1_20 = shiftPix+normUV* texture2D(TI16_2, iP2_0).ba; // P2index[iP2[0]][ia+1];
  cP1_20 = texture2D(TF32_scales, iP2_0).a * wa;         // P2scale[iP2[0]][ia+1]*wa;

  iP2_1.x += aux;
  iP1_11 = shiftPix+normUV* texture2D(TI16_2, iP2_1).ba; // P2index[iP2[1]][ia];
  cP1_11 = texture2D(TF32_scales, iP2_1).a * i_wa;       // P2scale[iP2[1]][ia]*1_wa;
  iP2_1.x += pixStep;
  iP1_21 = shiftPix+normUV* texture2D(TI16_2, iP2_1).ba; // P2index[iP2[1]][ia+1];
  cP1_21 = texture2D(TF32_scales, iP2_1).a * wa;         // P2scale[iP2[1]][ia+1]*wa;

  iP2_2.x += aux;
  iP1_12 = shiftPix+normUV* texture2D(TI16_2, iP2_2).ba; // P2index[iP2[2]][ia];
  cP1_12 = texture2D(TF32_scales, iP2_2).a * i_wa;       // P2scale[iP2[2]][ia]*1_wa;
  iP2_2.x += pixStep;
  iP1_22 = shiftPix+normUV* texture2D(TI16_2, iP2_2).ba; // P2index[iP2[2]][ia+1];
  cP1_22 = texture2D(TF32_scales, iP2_2).a * wa;         // P2scale[iP2[2]][ia+1]*wa;

  iP2_3.x += aux;
  iP1_13 = shiftPix+normUV* texture2D(TI16_2, iP2_3).ba; // P2index[iP2[3]][ia];
  cP1_13 = texture2D(TF32_scales, iP2_3).a * i_wa;       // P2scale[iP2[3]][ia]*1_wa;
  iP2_3.x += pixStep;
  iP1_23 = shiftPix+normUV* texture2D(TI16_2, iP2_3).ba; // P2index[iP2[3]][ia+1];
  cP1_23 = texture2D(TF32_scales, iP2_3).a * wa;         // P2scale[iP2[3]][ia+1]*wa;

  //I2 lookups =============================================
  vec2 iI1_10,iI1_20, iI1_11,iI1_21, iI1_12,iI1_22, iI1_13,iI1_23;

  iI2_0.x += aux;
  iI1_10 = shiftPix+normUV* texture2D(TI16_4, iI2_0).rg;  // I2index[iI2[0]][ia];
  iI2_0.x += pixStep;
  iI1_20 = shiftPix+normUV* texture2D(TI16_4, iI2_0).rg;  // I2index[iI2[0]][ia+1];

  iI2_1.x += aux;
  iI1_11 = shiftPix+normUV* texture2D(TI16_4, iI2_1).rg;  // I2index[iI2[1]][ia];
  iI2_1.x += pixStep;
  iI1_21 = shiftPix+normUV* texture2D(TI16_4, iI2_1).rg;  // I2index[iI2[1]][ia+1];

  iI2_2.x += aux;
  iI1_12 = shiftPix+normUV* texture2D(TI16_4, iI2_2).rg;  // I2index[iI2[2]][ia];
  iI2_2.x += pixStep;
  iI1_22 = shiftPix+normUV* texture2D(TI16_4, iI2_2).rg;  // I2index[iI2[2]][ia+1];

  iI2_3.x += aux;
  iI1_13 = shiftPix+normUV* texture2D(TI16_4, iI2_3).rg;  // I2index[iI2[3]][ia];
  iI2_3.x += pixStep;
  iI1_23 = shiftPix+normUV* texture2D(TI16_4, iI2_3).rg;  // I2index[iI2[3]][ia+1];

  //####################
  // DATA RECONSTRUCTION
  //####################
  aux = ib*pixStep;

  // I1 lookups =========================================
  vec2 valI1_0,valI1_1,valI1_2,valI1_3,valI1_4,valI1_5,valI1_6,valI1_7; 
  vec2 valI1_8,valI1_9,valI1_10,valI1_11,valI1_12,valI1_13,valI1_14,valI1_15; 

  iI1_10.x += aux;
  valI1_0 = shiftPix+normUV* texture2D(TI16_4, iI1_10).ba; // I1index[iI1_1[0]][ib];
  iI1_10.x += pixStep;
  valI1_1 = shiftPix+normUV* texture2D(TI16_4, iI1_10).ba; // I1index[iI1_1[0]][ib+1];
  
  iI1_20.x += aux;
  valI1_2 = shiftPix+normUV* texture2D(TI16_4, iI1_20).ba; // I1index[iI1_2[0]][ib];
  iI1_20.x += pixStep;
  valI1_3 = shiftPix+normUV* texture2D(TI16_4, iI1_20).ba; // I1index[iI1_2[0]][ib+1];
  
  iI1_11.x += aux;
  valI1_4 = shiftPix+normUV* texture2D(TI16_4, iI1_11).ba; // I1index[iI1_1[1]][ib];
  iI1_11.x += pixStep;
  valI1_5 = shiftPix+normUV* texture2D(TI16_4, iI1_11).ba; // I1index[iI1_1[1]][ib+1];

  iI1_21.x += aux;
  valI1_6 = shiftPix+normUV* texture2D(TI16_4, iI1_21).ba; // I1index[iI1_2[1]][ib];
  iI1_21.x += pixStep;
  valI1_7 = shiftPix+normUV* texture2D(TI16_4, iI1_21).ba; // I1index[iI1_2[1]][ib+1];

  iI1_12.x += aux;
  valI1_8 = shiftPix+normUV* texture2D(TI16_4, iI1_12).ba; // I1index[iI1_1[2]][ib];
  iI1_12.x += pixStep;
  valI1_9 = shiftPix+normUV* texture2D(TI16_4, iI1_12).ba; // I1index[iI1_1[2]][ib+1];

  iI1_22.x += aux;
  valI1_10= shiftPix+normUV* texture2D(TI16_4, iI1_22).ba; // I1index[iI1_2[2]][ib];
  iI1_22.x += pixStep;
  valI1_11= shiftPix+normUV* texture2D(TI16_4, iI1_22).ba; // I1index[iI1_2[2]][ib+1];

  iI1_13.x += aux;
  valI1_12= shiftPix+normUV* texture2D(TI16_4, iI1_13).ba; // I1index[iI1_1[3]][ib];
  iI1_13.x += pixStep;
  valI1_13= shiftPix+normUV* texture2D(TI16_4, iI1_13).ba; // I1index[iI1_1[3]][ib+1];

  iI1_23.x += aux;
  valI1_14= shiftPix+normUV* texture2D(TI16_4, iI1_23).ba; // I1index[iI1_2[3]][ib];
  iI1_23.x += pixStep;
  valI1_15= shiftPix+normUV* texture2D(TI16_4, iI1_23).ba; // I1index[iI1_2[3]][ib+1];

  // C lookups & interpolation =========================================
  vec2 valC;
  vec4 valA1,valB1, valA2,valB2;

  valC.x = texture2D(TF32_data, valI1_0).g;
  valC.y = texture2D(TF32_data, valI1_1).g;
  valA1[0] = mix(valC.x,valC.y,wb);          // 1_wb * valC.x + wb * valC.y;
  valI1_0.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_0).g;
  valI1_1.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_1).g;
  valB1[0] = mix(valC.x,valC.y,wb);          // 1_wb * valC.x + wb * valC.y;

  valC.x = texture2D(TF32_data, valI1_2).g;
  valC.y = texture2D(TF32_data, valI1_3).g;
  valA2[0] = mix(valC.x,valC.y,wb);          
  valI1_2.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_2).g;
  valI1_3.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_3).g;
  valB2[0] = mix(valC.x,valC.y,wb);          

  valC.x = texture2D(TF32_data, valI1_4).g;
  valC.y = texture2D(TF32_data, valI1_5).g;
  valA1[1] = mix(valC.x,valC.y,wb);          
  valI1_4.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_4).g;
  valI1_5.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_5).g;
  valB1[1] = mix(valC.x,valC.y,wb);          

  valC.x = texture2D(TF32_data, valI1_6).g;
  valC.y = texture2D(TF32_data, valI1_7).g;
  valA2[1] = mix(valC.x,valC.y,wb);          
  valI1_6.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_6).g;
  valI1_7.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_7).g;
  valB2[1] = mix(valC.x,valC.y,wb);          

  valC.x = texture2D(TF32_data, valI1_8).g;
  valC.y = texture2D(TF32_data, valI1_9).g;
  valA1[2] = mix(valC.x,valC.y,wb);          
  valI1_8.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_8).g;
  valI1_9.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_9).g;
  valB1[2] = mix(valC.x,valC.y,wb);          

  valC.x = texture2D(TF32_data, valI1_10).g;
  valC.y = texture2D(TF32_data, valI1_11).g;
  valA2[2] = mix(valC.x,valC.y,wb);          
  valI1_10.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_10).g;
  valI1_11.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_11).g;
  valB2[2] = mix(valC.x,valC.y,wb);          
   
  valC.x = texture2D(TF32_data, valI1_12).g;
  valC.y = texture2D(TF32_data, valI1_13).g;
  valA1[3] = mix(valC.x,valC.y,wb);          
  valI1_12.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_12).g;
  valI1_13.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_13).g;
  valB1[3] = mix(valC.x,valC.y,wb);          

  valC.x = texture2D(TF32_data, valI1_14).g;
  valC.y = texture2D(TF32_data, valI1_15).g;
  valA2[3] = mix(valC.x,valC.y,wb);          
  valI1_14.x += pixStep;
  valC.x = texture2D(TF32_data, valI1_14).g;
  valI1_15.x += pixStep;
  valC.y = texture2D(TF32_data, valI1_15).g;
  valB2[3] = mix(valC.x,valC.y,wb);          
  
  // I2 interpol. =========================================
  vec4 a, b;
  a[0] = mix(valA1[0],valA2[0],wa);         // valA1[0]*1_wa + valA2[0]*wa;
  b[0] = mix(valB1[0],valB2[0],wa);         // valB1[0]*1_wa + valB2[0]*wa;
  a[1] = mix(valA1[1],valA2[1],wa);         // valA1[1]*1_wa + valA2[1]*wa;
  b[1] = mix(valB1[1],valB2[1],wa);         // valB1[1]*1_wa + valB2[1]*wa;
  a[2] = mix(valA1[2],valA2[2],wa);         // valA1[2]*1_wa + valA2[2]*wa;
  b[2] = mix(valB1[2],valB2[2],wa);         // valB1[2]*1_wa + valB2[2]*wa;
  a[3] = mix(valA1[3],valA2[3],wa);         // valA1[3]*1_wa + valA2[3]*wa;
  b[3] = mix(valB1[3],valB2[3],wa);         // valB1[3]*1_wa + valB2[3]*wa;

  // P1 lookups & interpol. ==========================================
  vec2 val;
  vec4 valL1,valL2; 
  
  iP1_10.x += aux;
  val.x = texture2D(TF32_data, iP1_10).r;             // P1data[iP1_1[0]][ib];
  iP1_10.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_10).r;             // P1data[iP1_1[0]][ib+1];
  valL1[0] = mix(val.x,val.y,wb);                     // 1_wb * val.x + wb * val.y;

  iP1_20.x += aux;
  val.x = texture2D(TF32_data, iP1_20).r;             // P1data[iP1_2[0]][ib];
  iP1_20.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_20).r;             // P1data[iP1_2[0]][ib+1];
  valL2[0] = mix(val.x,val.y,wb);                     
  
  iP1_11.x += aux;
  val.x = texture2D(TF32_data, iP1_11).r;             // P1data[iP1_1[1]][ib];
  iP1_11.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_11).r;             // P1data[iP1_1[1]][ib+1];
  valL1[1] = mix(val.x,val.y,wb);                     

  iP1_21.x += aux;
  val.x = texture2D(TF32_data, iP1_21).r;             // P1data[iP1_2[1]][ib];
  iP1_21.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_21).r;             // P1data[iP1_2[1]][ib+1];
  valL2[1] = mix(val.x,val.y,wb); 

  iP1_12.x += aux;
  val.x = texture2D(TF32_data, iP1_12).r;             // P1data[iP1_1[2]][ib];
  iP1_12.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_12).r;             // P1data[iP1_1[2]][ib+1];
  valL1[2] = mix(val.x,val.y,wb); 

  iP1_22.x += aux;
  val.x = texture2D(TF32_data, iP1_22).r;             // P1data[iP1_2[2]][ib];
  iP1_22.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_22).r;             // P1data[iP1_2[2]][ib+1];
  valL2[2] = mix(val.x,val.y,wb); 

  iP1_13.x += aux;
  val.x = texture2D(TF32_data, iP1_13).r;             // P1data[iP1_1[3]][ib];
  iP1_13.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_13).r;             // P1data[iP1_1[3]][ib+1];
  valL1[3] = mix(val.x,val.y,wb); 

  iP1_23.x += aux;
  val.x = texture2D(TF32_data, iP1_23).r;             // P1data[iP1_2[3]][ib];
  iP1_23.x += pixStep;                  
  val.y = texture2D(TF32_data, iP1_23).r;             // P1data[iP1_2[3]][ib+1];
  valL2[3] = mix(val.x,val.y,wb); 

  // PDF2D interpol. =======================================
  //  a = vec4(0.0, 0.0, 0.0, 0.0);   // for Luminance channel using only 
  //  b = vec4(0.0, 0.0, 0.0, 0.0);   // 
  vec4 L;
  L[0] = cP1_10*valL1[0] + cP1_20*valL2[0];
  L[1] = cP1_11*valL1[1] + cP1_21*valL2[1];
  L[2] = cP1_12*valL1[2] + cP1_22*valL2[2];
  L[3] = cP1_13*valL1[3] + cP1_23*valL2[3];

  // Conversion to RGB ===================================
  vec3 RGB1,RGB2,RGB3,RGB4;
  if (hdrMultiplier == 0.0) {
    // LDR data
    RGB1 = YCbCrtoRGB(L[0],a[0],b[0]);
    RGB2 = YCbCrtoRGB(L[1],a[1],b[1]);
    RGB3 = YCbCrtoRGB(L[2],a[2],b[2]);
    RGB4 = YCbCrtoRGB(L[3],a[3],b[3]);
  }
  else {
    // HDR data
    RGB1 = LogLuv2RGB(L[0],a[0],b[0]);
    RGB2 = LogLuv2RGB(L[1],a[1],b[1]);
    RGB3 = LogLuv2RGB(L[2],a[2],b[2]);
    RGB4 = LogLuv2RGB(L[3],a[3],b[3]);
  }

  // PDF3D interpol. =======================================
  vec2 texelR,texelG,texelB;
  texelR[0] = cM_11*RGB1.x + cM_21*RGB2.x;   
  texelR[1] = cM_12*RGB3.x + cM_22*RGB4.x;   

  texelG[0] = cM_11*RGB1.y + cM_21*RGB2.y;   
  texelG[1] = cM_12*RGB3.y + cM_22*RGB4.y;   

  texelB[0] = cM_11*RGB1.z + cM_21*RGB2.z;   
  texelB[1] = cM_12*RGB3.z + cM_22*RGB4.z;   

  // PDF4D, PDF6D interpol. ================================
  vec4 texel;

  texel.r = cP4*(cP3_1*texelR[0] + cP3_2*texelR[1]); 
  texel.g = cP4*(cP3_1*texelG[0] + cP3_2*texelG[1]); 
  texel.b = cP4*(cP3_1*texelB[0] + cP3_2*texelB[1]); 
  texel.a = 1.0;

  gl_FragColor = texel;

 }
}
