/* 
Name: OBJloader.cpp
Version: 1.0

Description: Source file for Wawefront OBJ file geometry loading
Author: Jiri Filip, UTIA AS CR, filipj'at'utia'dot'cas'dot'cz
Date: 28/05/07

COPYRIGHT ======================================================================

This source code is release under Creative Commons licence
(Attribution-NonCommercial 3.0 Unported) specified here:

http://creativecommons.org/licenses/by-nc/3.0/
http://creativecommons.org/licenses/by-nc/3.0/legalcode

You are free:
    * to Share - to copy, distribute and transmit the work
    * to Remix - to adapt the work
Under the following conditions:
    * Attribution - You must attribute the work in the manner specified by the
      author or licensor (but not in any way that suggests that they endorse you
      or your use of the work).

      Attribute this work:
      ------------------------------------------------
      Bidirectional Texture Function Compression based on Multi-Level Vector 
      Quantization
      as the paper written by
          Vlastimil Havran 	Jiri Filip 	Karol Myszkowski

      and published in Computer Graphics Forum, Volume 29, Issue 1, 
      pages 175-190, March 2010, DOI: 10.1111/j.1467-8659.2009.01585.x

      Project page:
        http://www.cgg.cvut.cz/members/havran/btfbase/
      ------------------------------------------------
     
    * Noncommercial - You may not use this work for commercial purposes.

With the understanding that:
    * Waiver - Any of the above conditions can be waived if you get permission 
      from the copyright holder.
    * Public Domain - Where the work or any of its elements is in the public 
      domain under applicable law, that status is in no way affected by 
      the license.
    * Other Rights - In no way are any of the following rights affected by the 
      license:
          o Your fair dealing or fair use rights, or other applicable copyright 
            exceptions and limitations;
          o The author's moral rights;
          o Rights other persons may have either in the work itself or in how 
            the work is used, such as publicity or privacy rights.
    * Notice - For any reuse or distribution, you must make clear to others 
         the license terms of this work. The best way to do this is with a link 
         to this web page.
===========================================================================  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cassert>
#include <float.h> // for DOT. NET _isnan

#include "TAlloc.h"
#include "OBJloader.h"

//########################################################################################
//#   TObjectData                                                                        #
//########################################################################################

TObjectData::TObjectData(int dataSize)
{
  this->dataSize = dataSize;
  fileName = callocation1(0,1000);
  OBJfileName = callocation1(0,1000);
  vertices = allocation2(1,dataSize,0,3);
  UVcoord = allocation2(1,dataSize,0,3);
  normals = allocation2(1,dataSize,0,3);
  faces = lallocation2(1,dataSize,0,8);

  tangentsF = NULL; 
  tangentsV = NULL; 
  normalsF = NULL; 
  normalsV = NULL; 
}//--- TObjectData --------------------------------------------------------

TObjectData::~TObjectData()
{
  freecmem1(OBJfileName, 0,1000);
  freecmem1(fileName, 0,1000);  

  freemem2(vertices,1,dataSize,0,3);
  freemem2(UVcoord,1,dataSize,0,3);
  freemem2(normals,1,dataSize,0,3);
  freelmem2(faces,1,dataSize,0,8);

  if(tangentsF!=NULL)
    freemem2(tangentsF,1,dataSize,0,2); 
  if(tangentsV!=NULL)
    freemem2(tangentsV,1,dataSize,0,2); 
  if(normalsF!=NULL)
    freemem2(normalsF,1,dataSize,0,2); 
  if(normalsV!=NULL)
    freemem2(normalsV,1,dataSize,0,2); 
}//--- TObjectData --------------------------------------------------------

int TObjectData::getNoOfTokens(char *line, char *delimiter)
{
  char *token;

  int NoOfTokens=0;
  token = strtok(line,delimiter);
  while(token!=NULL)
    {
      NoOfTokens++;
      token = strtok(NULL,delimiter);
    }
  return NoOfTokens;
}//--- getNoOfTokens -----------------------------------------------------

int TObjectData::loadOBJfile(char *_fileName, int center, float scale, float xoff, float yoff, float zoff)
{
  strcpy(OBJfileName, _fileName);

  printf("Loading 3D geometry from file '%s'...",_fileName);fflush(NULL);
  FILE *fp;

  if((fp=fopen(_fileName,"r"))==NULL)
    {
      printf("Filename '%s' is wrong or file doesn't exist\n",_fileName);
      exit(0);
    }

  i_v=1; i_vt=1; i_vn=1; i_f=1;
  char dataType[100];
  float v[3], vt[3];
  int f[3], ft[3];
  int assignedItems;
  int NoOfTokens;
  char dt[100];
  char line2[1000];
  int lineIndex=0;
  long int polygon[4][3];

  char *line = callocation1(0,1000);
  char cd;
  while(fgets(line,1000,fp)!=NULL) {
    lineIndex++;
    assignedItems = sscanf(line,"%s",&dataType);
    strcpy(line2,line);
    NoOfTokens = getNoOfTokens(line," ");

    if(assignedItems == 1) // if not empty line
      {        
        if(strcmp(dataType,"v")==0)
          if(NoOfTokens==4)
            {
              sscanf(line2,"%s %f %f %f",&dt, &vertices[i_v][0],&vertices[i_v][1],&vertices[i_v][2]);
              i_v++;
            }
          else
            {
              printf("  Line: %d - Wrong number of vertex coordinates %d (expected 3)!!! Exiting...\n",lineIndex,NoOfTokens-1);
              exit(-1);
            }

        if(strcmp(dataType,"vt")==0)
          {
            if(NoOfTokens==3)
              {
                sscanf(line2,"%s %f %f\n",&dt, &UVcoord[i_vt][0],&UVcoord[i_vt][1]);
                i_vt++;
              }
            else if(NoOfTokens==4)
              {
                sscanf(line2,"%s %f %f %f\n",&dt, &UVcoord[i_vt][0],&UVcoord[i_vt][1],&UVcoord[i_vt][2]);
                i_vt++;
              }
            else
              {
                printf("  Line: %d - Wrong number of UV coordinates %d (expected 2 or 3)!!! Exiting...\n",lineIndex,NoOfTokens-1);
                exit(-1);
              }
          }

        if(strcmp(dataType,"vn")==0)
          if(NoOfTokens==4)
            {
              sscanf(line2,"%s %f %f %f\n",&dt, &normals[i_vn][0],&normals[i_vn][1],&normals[i_vn][2]);
              i_vn++;
            }
          else
            {
              printf("  Line: %d - Wrong number of normal coordinates %d (expected: 3)!!! Exiting...\n",lineIndex,NoOfTokens-1);
              exit(-1);
            }

        if(strcmp(dataType,"f")==0)
          {
            if(NoOfTokens==4)
              {
                if((i_vn>1)&&(i_vt>1))
                  sscanf(line2,"%s %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",&dt, &faces[i_f][0],&faces[i_f][3],&faces[i_f][6],&faces[i_f][1],&faces[i_f][4],&faces[i_f][7],&faces[i_f][2],&faces[i_f][5],&faces[i_f][8]);
                // f v//vn
                if((i_vn>1)&&(i_vt==1))
                  sscanf(line2,"%s %ld//%ld %ld//%ld %ld//%ld\n",&dt, &faces[i_f][0],&faces[i_f][6],&faces[i_f][1],&faces[i_f][7],&faces[i_f][2],&faces[i_f][8]);
                  
                // f v/vt
                if((i_vn==1)&&(i_vt>1))
                  sscanf(line2,"%s %ld/%ld %ld/%ld %ld/%ld\n",&dt, &faces[i_f][0],&faces[i_f][3],&faces[i_f][1],&faces[i_f][4],&faces[i_f][2],&faces[i_f][5]);
                i_f++;
              }
            else if(NoOfTokens==5)
              {
                if((i_vn>1)&&(i_vt>1))
                  {
                    sscanf(line2,"%s %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld %ld/%ld/%ld\n",&dt, &polygon[0][0],&polygon[0][1],&polygon[0][2],  &polygon[1][0],&polygon[1][1],&polygon[1][2],  &polygon[2][0],&polygon[2][1],&polygon[2][2],  &polygon[3][0],&polygon[3][1],&polygon[3][2]);
                    // first triangle
                    // vertices
                    faces[i_f][0] = polygon[0][0];
                    faces[i_f][1] = polygon[1][0];
                    faces[i_f][2] = polygon[2][0];
                    // UV coords.
                    faces[i_f][3] = polygon[0][1];
                    faces[i_f][4] = polygon[1][1];
                    faces[i_f][5] = polygon[2][1];
                    // normals
                    faces[i_f][6] = polygon[0][2];
                    faces[i_f][7] = polygon[1][2];
                    faces[i_f][8] = polygon[2][2];
                    i_f++;
                    // second triangle
                    // vertices
                    faces[i_f][0] = polygon[3][0];
                    faces[i_f][1] = polygon[0][0];
                    faces[i_f][2] = polygon[2][0];
                    // UV coords.
                    faces[i_f][3] = polygon[3][1];
                    faces[i_f][4] = polygon[0][1];
                    faces[i_f][5] = polygon[2][1];
                    // normals
                    faces[i_f][6] = polygon[3][2];
                    faces[i_f][7] = polygon[0][2];
                    faces[i_f][8] = polygon[2][2];
                    i_f++;                    
                  }
                  
                // f v//vn
                if((i_vn>1)&&(i_vt==1))
                  {
                    sscanf(line2,"%s %ld//%ld %ld//%ld %ld//%ld %ld//%ld\n",&dt, &polygon[0][0],&polygon[0][2],&polygon[1][0],&polygon[1][2],&polygon[2][0],&polygon[2][2],&polygon[3][0],&polygon[3][2]);
                    // first triangle
                    // vertices
                    faces[i_f][0] = polygon[0][0];
                    faces[i_f][1] = polygon[1][0];
                    faces[i_f][2] = polygon[2][0];
                    // normals
                    faces[i_f][6] = polygon[0][2];
                    faces[i_f][7] = polygon[1][2];
                    faces[i_f][8] = polygon[2][2];
                    i_f++;
                    // second triangle
                    // vertices
                    faces[i_f][0] = polygon[3][0];
                    faces[i_f][1] = polygon[0][0];
                    faces[i_f][2] = polygon[2][0];
                    // normals
                    faces[i_f][6] = polygon[3][2];
                    faces[i_f][7] = polygon[0][2];
                    faces[i_f][8] = polygon[2][2];
                    i_f++;                    
                  }
                  
                // f v/vt
                if((i_vn==1)&&(i_vt>1))
                  {
                    sscanf(line2,"%s %ld/%ld %ld/%ld %ld/%ld %ld/%ld\n",&dt, &polygon[0][0],&polygon[0][1],&polygon[1][0],&polygon[1][1],&polygon[2][0],&polygon[2][1],&polygon[3][0],&polygon[3][1]);
                    // first triangle
                    // vertices
                    faces[i_f][0] = polygon[0][0];
                    faces[i_f][1] = polygon[1][0];
                    faces[i_f][2] = polygon[2][0];
                    // UV coords.
                    faces[i_f][3] = polygon[0][1];
                    faces[i_f][4] = polygon[1][1];
                    faces[i_f][5] = polygon[2][1];
                    i_f++;
                    // second triangle
                    // vertices
                    faces[i_f][0] = polygon[3][0];
                    faces[i_f][1] = polygon[0][0];
                    faces[i_f][2] = polygon[2][0];
                    // UV coords.
                    faces[i_f][3] = polygon[3][1];
                    faces[i_f][4] = polygon[0][1];
                    faces[i_f][5] = polygon[2][1];
                    i_f++;                    
                  }
              }
            else
            {
              printf("  Line: %d - Wrong number of polygon vertices %d (expected: 3 or 4)!!! Exiting...\n",lineIndex,NoOfTokens-1);
              exit(-1);
            }

          }// f
      }
  }// while not final line

  printf("done\n");
  printf("   Total of:\n        %10d vertices\n        %10d UV coordinates\n        %10d normals\n        %10d faces               were loaded\n\n",i_v-1,i_vt-1,i_vn-1,i_f-1);
  fclose(fp);
  // vertices centering and scaling
  long i;
  double sum[3];
  if(center)
    {
      sum[0] = sum[1] = sum[2] = 0.0;
      for(i=1;i<i_v;i++)
        {
          sum[0] += vertices[i][0];
          sum[1] += vertices[i][1];
          sum[2] += vertices[i][2];
        }
      for(i=0;i<3;i++)
        sum[i] /= i_v-1;
      printf("   Object Center: %3.2f %3.2f %3.2f\n",sum[0],sum[1],sum[2]);
      for(i=1;i<i_v;i++)
        {
          vertices[i][0] -= sum[0];
          vertices[i][1] -= sum[1];
          vertices[i][2] -= sum[2];
        }
    }

  for(i=1;i<i_v;i++)
    {
      vertices[i][0] -= xoff;
      vertices[i][1] -= yoff;
      vertices[i][2] -= zoff;
    }

  // scaling vertices
  for(i=1;i<i_v;i++)
    {
      vertices[i][0] *= scale;
      vertices[i][1] *= scale;
      vertices[i][2] *= scale;
    }

  // checking UV coords if correction needed
  float Xmax=-1e10,Xmin=1e10;
  float Ymax=-1e10,Ymin=1e10;
  for(i=1;i<i_vt;i++)
    {
      if(UVcoord[i][0]>Xmax)
        Xmax = UVcoord[i][0];
      if(UVcoord[i][0]<Xmin)
        Xmin = UVcoord[i][0];
      if(UVcoord[i][1]>Ymax)
        Ymax = UVcoord[i][1];
      if(UVcoord[i][1]<Ymin)
        Ymin = UVcoord[i][1];
    }
  for(i=1;i<i_vt;i++)
    {
      UVcoord[i][0] = (UVcoord[i][0]-Xmin)/(Xmax-Xmin); 
      UVcoord[i][1] = (UVcoord[i][1]-Ymin)/(Ymax-Ymin); 
    }

  // nomals normalization
  for(i=1;i<i_vn;i++)
    vec_norm(normals[i]);

  freecmem1(line, 0,1000);

  return 1;
}//--- loadOBJfile -----------------------------------------------------------------

long TObjectData::noOfFaces()
{
  return (i_f-1);
}//--- noOfFaces -----------------------------------------------------------------------

long TObjectData::noOfVertices()
{
  return (i_vt-1);
}//--- noOfVertices -----------------------------------------------------------------------

long TObjectData::noOfNormals()
{
  return (i_vn-1);
}//--- noOfNormalss -----------------------------------------------------------------------

void TObjectData::tangentsPerFace()
{
  if(i_vt==1)
    {
      printf("OBJloader: !!! UV coordinates are not available in the model !!! Exiting...\n");
      exit(-1);
    }
  printf("Computing tangents for individual faces...\r");fflush(stdout);
  if(tangentsF==NULL)
    tangentsF = allocation2(1,dataSize,0,2); 
  else
    return;

  float den;
  float **UV = allocation2(0,2, 0,2); 
  float **v = allocation2(0,2, 0,2); 
  float **n = allocation2(0,2, 0,2); 

  // computing tangents for individual faces
  for(int i=1;i<i_f;i++)
    {
      //for degenerated triangles skip computations
      if(faces[i][0]==faces[i][1] || faces[i][0]==faces[i][2] || faces[i][2]==faces[i][1])
        {
          for(int j=0;j<=2;j++)
            tangentsF[i][j] = 0.f;
          continue;
        }
      for(int j=0;j<=2;j++)
        {
          v[0][j] = vertices[faces[i][0]][j];
          v[1][j] = vertices[faces[i][1]][j];
          v[2][j] = vertices[faces[i][2]][j];
        }        

      UV[0][0] = UVcoord[faces[i][3]][0];
      UV[0][1] = UVcoord[faces[i][3]][1];
      UV[1][0] = UVcoord[faces[i][4]][0];
      UV[1][1] = UVcoord[faces[i][4]][1];
      UV[2][0] = UVcoord[faces[i][5]][0];
      UV[2][1] = UVcoord[faces[i][5]][1];


      den = UV[2][0]*(UV[1][1]-UV[0][1]) 
          - UV[1][0]*(UV[2][1]-UV[0][1]) 
          - UV[0][0]*(UV[1][1]-UV[2][1]);  

			if(fabs(den)>1e-15) {
      tangentsF[i][0] = (v[2][0]*(UV[1][1]-UV[0][1]) 
                       - v[1][0]*(UV[2][1]-UV[0][1]) 
                       - v[0][0]*(UV[1][1]-UV[2][1])) / den;
      tangentsF[i][1] = (v[2][1]*(UV[1][1]-UV[0][1]) 
                       - v[1][1]*(UV[2][1]-UV[0][1]) 
                       - v[0][1]*(UV[1][1]-UV[2][1])) / den;
      tangentsF[i][2] = (v[2][2]*(UV[1][1]-UV[0][1]) 
                       - v[1][2]*(UV[2][1]-UV[0][1]) 
                       - v[0][2]*(UV[1][1]-UV[2][1])) / den;
      vec_norm(tangentsF[i]);
      }
      else
        for(int j=0;j<=2;j++)
          tangentsF[i][j] = 0.1f;

      if(vec_len(tangentsF[i])<=0.f) 
        {
          printf("\ntangentsPerFace: Zero tangent - [%g,%g,%g]!!! Exiting...\n",tangentsF[i][0],tangentsF[i][1],tangentsF[i][2]);
          exit(-1);
        }

      if(_isnan(tangentsF[i][0]) || _isnan(tangentsF[i][0]) || _isnan(tangentsF[i][0]))
        {
          printf("\ntangentsPerFace: NaN in tangent at face: %d - [%g,%g,%g]. Exiting...\n",i,tangentsF[i][0],tangentsF[i][1],tangentsF[i][2]);
          printf("                      UV0   - [%g,%g]\n",UV[0][0],UV[0][1]);
          printf("                      UV1   - [%g,%g]\n",UV[1][0],UV[1][1]);
          printf("                      UV2   - [%g,%g]\n\n",UV[2][0],UV[2][1]);
          printf("                      V0   - [%g,%g,%g]\n",v[0][0],v[0][1],v[0][2]);
          printf("                      V1   - [%g,%g,%g]\n",v[1][0],v[1][1],v[1][2]);
          printf("                      V2   - [%g,%g,%g]\n",v[2][0],v[2][1],v[2][2]);
          exit(-1);
        }
    }// for faces
  printf("Computing tangents for individual faces...done\n");fflush(stdout);
  freemem2(UV, 0,2, 0,2); 
  freemem2(v,  0,2, 0,2); 
  freemem2(n,  0,2, 0,2); 
  return;
}//--- tangentsPerFace ---------------------------------------------------------------

void TObjectData::tangentsPerVertex()
{
  printf("Computing tangents for individual vertices...      \r");fflush(stdout);
  if(tangentsV==NULL)
    tangentsV = allocation2(1,dataSize,0,2); 
  else
    return;

  strcpy(fileName,OBJfileName);
  fileName[strlen(OBJfileName)-4] = '\0';
  strcat(fileName,".tan");

  bool status = loadTangentsPerVertex(fileName);
  printf("Loading status: '%s' (%d)\n",fileName, status);
  if(status == false)
    {
      if(tangentsF==NULL)
        this->tangentsPerFace();
      // computing tangents for individual vertices
      int trianglesPerVertex;
      float v[3][3];
      for(int i=1;i<i_v;i++)
        {
          for(int j=0;j<=2;j++)
            tangentsV[i][j] = 0.f;
          trianglesPerVertex = 0;
          for(int k=1;k<i_f;k++)
            {
              
              for(int j=0;j<=2;j++)
                {
                  v[0][j] = vertices[faces[k][0]][j];
                  v[1][j] = vertices[faces[k][1]][j];
                  v[2][j] = vertices[faces[k][2]][j];
                }
              if( ((v[0][0] == vertices[i][0])&&(v[0][1] == vertices[i][1])&&(v[0][2] == vertices[i][2]))||
                  ((v[1][0] == vertices[i][0])&&(v[1][1] == vertices[i][1])&&(v[1][2] == vertices[i][2]))||
                  ((v[2][0] == vertices[i][0])&&(v[2][1] == vertices[i][1])&&(v[2][2] == vertices[i][2])) )
                
                //                      if((i == faces[k][0])||(i == faces[k][1])||(i == faces[k][2]))
                {
                  for(int j=0;j<=2;j++)
                    tangentsV[i][j] += tangentsF[k][j];
                  trianglesPerVertex++;
                }
            }// for faces
          if(trianglesPerVertex >= 1)
            for(int j=0;j<=2;j++)
              tangentsV[i][j] /= trianglesPerVertex;
          else
            for(int j=0;j<=2;j++)
              tangentsV[i][j] = 0.1f;
          
          if(_isnan(tangentsV[i][0]) || _isnan(tangentsV[i][0]) || _isnan(tangentsV[i][0])) 
            {
              printf("\ntangentsPerVertex: NaN in tangent at face: %d - [%g,%g,%g] %d. Exiting...\n",i,tangentsV[i][0],tangentsV[i][1],tangentsV[i][2], trianglesPerVertex);
              exit(-1);
            }
          
          if(vec_len(tangentsV[i])<=0.f) 
            {
              printf("\ntangentsPerVertex: Zero tangent at face: %d - [%g,%g,%g]!!! Exiting...\n",i,tangentsV[i][0],tangentsV[i][1],tangentsV[i][2]);
              exit(-1);
            }
          vec_norm(tangentsV[i]);
        }// for vertices
      printf("Computing tangents for individual vertices...done  \n");fflush(stdout);
      
      // saving computed tangents (if not done already)
      printf("Saving tangents per vertex                   \r");
      saveTangentsPerVertex(fileName);
    }

  return;
}//--- tangentsPerVertex ----------------------------------------------------------------

void TObjectData::saveTangentsPerVertex(char *fileName)
{
  FILE *fp;
  
  printf("\nSaving: %s                                \n",fileName);

  if((fp=fopen(fileName,"w"))==NULL)
    {
      printf(" Error when creating file: '%s' !!!\n",fileName);
      exit(-1);
    }

  fprintf(fp,"%d\n",i_v-1);
  for(int i=1;i<i_v;i++)
    fprintf(fp,"%f %f %f\n",tangentsV[i][0],tangentsV[i][1],tangentsV[i][2]);
  
  fclose(fp);    
  printf(" Computed tangents per vertex were saved to file: '%s'\n",fileName);

  return;
}//--- saveTangentsPerVertex ------------------------------------------------------------

bool TObjectData::loadTangentsPerVertex(char *fileName)
{
  FILE *fp;
  
  printf("Loading: %s                                \n",fileName);

  if((fp=fopen(fileName,"r"))==NULL)
    {
      printf(" Cannot open file: '%s'. Computing tangents instead...\n",fileName);
      return false;
    }
  int NoOfTangents;
  fscanf(fp,"%d\n",&NoOfTangents);
  if(NoOfTangents > i_v-1) { 
    printf("Wrong number of tangents %d (assumed %d) !!! Exiting...\n",NoOfTangents, i_v);
    exit(-1);
  }
  for(int i=1;i<=NoOfTangents;i++)
    fscanf(fp,"%f %f %f\n",&tangentsV[i][0],&tangentsV[i][1],&tangentsV[i][2]);
  
  fclose(fp);    
  
  return true;
}//--- loadTangentsPerVertex ------------------------------------------------------------

float TObjectData::getV(int face, int vertex, int xyz)
{
  return vertices[faces[face][vertex]][xyz];
}//--- getV ----------------------------------------------------------------------------

int TObjectData::getVindex(int face, int vertex)
{
  return faces[face][vertex];
}//--- getVindex -----------------------------------------------------------------------

float TObjectData::getN(int face, int vertex, int xyz)
{
  if(i_vn>1)
    return normals[faces[face][6+vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Normals per vertex are not available in the model !!! Exiting...\n");
      exit(-1);
    }
}//--- getN ----------------------------------------------------------------------------

float TObjectData::getNv(int face, int vertex, int xyz)
{
  if(normalsV!=NULL)
    return normalsV[faces[face][6+vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Normals per vertex are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getNv ----------------------------------------------------------------------------

float TObjectData::getNf(int face, int xyz)
{
  if(normalsF!=NULL)
    return normalsF[face][xyz];
  else
    {
      printf("OBJloader: !!! Normals per face are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getNf ---------------------------------------------------------------------------

float TObjectData::getT(int face, int vertex, int xyz)
{
  if(tangentsV!=NULL)
    return tangentsV[faces[face][vertex]][xyz];
  else
    {
      printf("OBJloader: !!! Tangents per vertex are not computed !!! Exiting...\n");
      exit(-1);
    }

}//--- getT ----------------------------------------------------------------------------

float TObjectData::getTf(int face, int xyz)
{
  if(tangentsF!=NULL)
  return tangentsF[face][xyz];
  else
    {
      printf("OBJloader: !!! Tangents per face are not computed !!! Exiting...\n");
      exit(-1);
    }
}//--- getTf ---------------------------------------------------------------------------

float TObjectData::getUV(int face, int vertex, int uv)
{
  if(i_vt>1)
    return UVcoord[faces[face][3+vertex]][uv];
  else
    {
      printf("OBJloader: !!! UV coordinates are not available in the model !!! Exiting...\n");
      exit(-1);
    }
}//--- getUV ----------------------------------------------------------------------------


//###############################################################
// 3D Vector operations
//###############################################################

void TObjectData::vec_crossprod(float *x, float *y, float *res)
{
    res[0] = x[1]*y[2] - x[2]*y[1];
    res[1] = x[2]*y[0] - x[0]*y[2];
    res[2] = x[0]*y[1] - x[1]*y[0];
}//--- vec_crossprod ---------------------------------------------

float TObjectData::vec_len(float *x)
{
    return sqrtf(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
}//--- vec_len ---------------------------------------------------

void TObjectData::vec_diff(float *x, float *y, float *res)
{
    res[0] = x[0] - y[0];
    res[1] = x[1] - y[1];
    res[2] = x[2] - y[2];
}//--- vec_diff ---------------------------------------------------

int TObjectData::vec_norm(float *x)
{
    float aux;
    aux = vec_len(x);
    if(aux == 0.f)
      {		
        x[0] = 0.f;
        x[1] = 0.f;
        x[2] = 0.f;
        printf("vec_norm: zero length vector");    
        return 0;
      }
    x[0] = x[0]/aux;
    x[1] = x[1]/aux;
    x[2] = x[2]/aux;
    return 1;
}//--- vec_norm ---------------------------------------------------

void TObjectData::normal(float *v1, float *v2, float *v3, float *n)
{
    float *p, *q;
    p = allocation1(0,2);
    q = allocation1(0,2);
    
    vec_diff(v1,v2,p);
    vec_diff(v3,v2,q);
    vec_crossprod(q,p,n);
    vec_norm(n);

    freemem1(p,0,2);
    freemem1(q,0,2);
}//--- normal -----------------------------------------------------
